# Copyright (c) 2017 David Preece - davep@polymath.tech, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import logging
import os
import shortuuid
from threading import Thread
from bottle import Bottle, run, request
from litecache.cache import SqlCache


ident_init = """
CREATE TABLE nodes (pk TEXT NOT NULL UNIQUE, json BLOB);
CREATE TABLE users (pk TEXT NOT NULL UNIQUE, email TEXT NOT NULL, json BLOB);
CREATE TABLE pending (token TEXT NOT NULL UNIQUE, email TEXT NOT NULL);
"""


class Identity:
    """A default provider of identity and configuration"""

    def __init__(self, directory="~"):
        """Construct the identity database if it's not there.

        :param directory: The directory in which to place the database (identity.sqlite3)"""
        self.db = SqlCache(os.path.expanduser(directory), 'identity', ident_init)

    def stop(self):
        self.db.close()

    def create_pending_user(self, email) -> str:
        """Registers the intention for someone to become a registered user

        :param email: email address of the user.
        :return: confirmation token to give to the user."""
        token = shortuuid.uuid()
        self.db.mutate("INSERT INTO pending (token, email) VALUES (?, ?)", (token, email))
        return token

    def pending_users_for_token(self, token) -> []:
        """Return the pending users for the given token (may well be zero).

        :param token: the token a user was given in order to be able to confirm their account.
        :return: The list of pending users for that token."""
        return self.db.query("SELECT email FROM pending WHERE token=?", (token,))

    def register_user(self, pk_b64: str, email: str, config: str):
        """Registers a user as being valid.

        :param pk_b64: The user's primary key - base64 encoded string.
        :param email: The user's email address.
        :param config: A json description of any configuration to be associated with the user."""
        self.db.mutate("DELETE FROM pending WHERE email=?", (email,))
        self.db.mutate("INSERT INTO users (pk, email, json) VALUES (?, ?, ?)", (pk_b64, email, config))

    def user_config_from_db(self, pk_b64: str) -> (str, str):  # is used to check for presence in the db, too
        """Returns the json configuration for a user.

        :param pk_b64: The user's primary key - base64 encoded string.
        :return: A tuple of email address and the json configuration."""
        return self.db.query_one("SELECT email, json FROM users WHERE pk=?", (pk_b64,), "Unknown user")

    def raise_for_no_user(self, email: str):
        """Raises an error if this email address does not have an account.

        :param email: email address of the user."""
        self.db.query_one("SELECT * FROM users WHERE email=?", (email,), "no validated account")

    def register_node(self, pk_b64: str, config: str):
        """Writes a node's configuration into the database.

        :param pk_b64: The node's primary key - base64 encoded string.
        :param config: A json description of any configuration to be associated with the node."""
        self.db.mutate("INSERT INTO nodes (pk, json) VALUES (?, ?)", (pk_b64, config))

    def node_config_from_db(self, pk_b64: str) -> str:
        """Returns the json configuration of a node.

        :param pk_b64: The node's primary key - base64 encoded string.
        :return: The json configuration for the node."""
        return self.db.query_one("SELECT json FROM nodes WHERE pk=?", (pk_b64,), "Unknown node")[0]


confirmation_server = Bottle()


class AccountConfirmationServer(Thread):
    """A simple HTTP server for confirming accounts"""
    # has single use tokens so no real need to SSL
    identity = None
    pk = None
    port = None

    def __init__(self, identity, keys, port):
        super().__init__(name=str("Account Confirmation Server"), daemon=True)
        AccountConfirmationServer.identity = identity
        AccountConfirmationServer.pk = keys.public
        AccountConfirmationServer.port = port
        self.start()

    @staticmethod
    @confirmation_server.route('/', method='POST')
    def account():
        # de-HTTP the request
        try:
            token, user_pk = request.body.read().decode().split()
        except:
            logging.warning("Off-spec request to account creation server: " + request.body.read().decode())
            return None

        # valid token?
        pending_records = AccountConfirmationServer.identity.pending_users_for_token(token)
        if len(pending_records) == 0:
            logging.warning("An attempt was made to confirm an account with an incorrect token: " + token)
            return "Fail: this token is either incorrect or has been used already."
        user_email = pending_records[0][0]

        # all good
        AccountConfirmationServer.identity.register_user(user_pk, user_email, "{}")
        logging.info("Confirmed an account for: " + user_email)
        return AccountConfirmationServer.pk

    def run(self):
        try:
            logging.info("Started account confirmation server: 0.0.0.0:" + str(AccountConfirmationServer.port))
            run(app=confirmation_server, host='0.0.0.0', port=AccountConfirmationServer.port, quiet=True)
        except OSError:
            logging.critical("Could not bind account confirmation server, exiting")
            exit(1)



# Copyright (c) 2017 David Preece - davep@polymath.tech, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import logging
import os
import libnacl
import libnacl.utils
import cbor
import psutil
import time
from threading import Thread
from _thread import allocate_lock, get_ident
from base64 import b64decode
import shortuuid
import zmq
from zmq.utils.monitor import recv_monitor_message
from .message import Message
from messidge import KeyPair, Waitable
from messidge.loop import Loop


class Connection(Waitable):
    """Connection onto Messidge broker."""

    # Is expecting ~/.messidge/ to contain keys named after the location to connect to (eg)
    # -rw-r--r--  1 dpreece  staff   45B Jul  8 16:36 localhost
    # -r--------  1 dpreece  staff   45B Jul  8 16:36 localhost.pub

    # skt is the main tcp socket from here to the location and belongs to the background thread
    # x_thread_socket is the "pickup" end of the per-thread sending sockets forwards to skt
    # Note that messages received are exclusively dealt with on the background thread

    def __init__(self, location: str=None, *, prefix='~/.messidge',
                 exit_on_exception=False, reflect_value_errors=False,
                 location_ip: str=None, keys: KeyPair=None, server_pk: str=None):
        """Instantiate a connection.

        :param location: The FQDN of the location to connect to.
        :param prefix: Directory for the client keys and server public keys.
        :param exit_on_exception: Causes the message loop to exit if there's an uncaught exception.
        :param reflect_value_errors: If True, raised ValueErrors will have their source messages replied to.
        :param location_ip: An override for the DNS resolution of 'location'. Useful for LAN and/or test connections.
        :param keys: An override for the key pair in the 'prefix' directory.
        :param server_pk: An override for the server public key in the 'prefix' directory."""
        super().__init__()
        self.connect_ip = location_ip if location_ip is not None else location
        self.location = location if location is not None else self.connect_ip
        self.keys = keys if keys is not None else KeyPair(location, prefix=prefix)
        self.inproc_name = "inproc://x_thread/" + str(id(self))
        self.session_key = None
        self.connect_callbacks = set()
        self.reflect_value_errors = reflect_value_errors
        self.exit_on_exception = exit_on_exception
        self.thread_skt = {}  # map from thread to the socket used to send messages on it's behalf
        self.rid = b''
        self.loop = None
        self.loop_block = allocate_lock()
        self.uuid_blockreply = {}
        self.uuid_blockresults = {}
        self.connected = False
        self.loop_thread = None

        # fetch the server's public key
        if server_pk is None:
            try:
                filename = os.path.expanduser(prefix + '/%s.spub' % location)
                with open(filename, 'r') as f:
                    server_pk = f.read().strip('\n')
            except FileNotFoundError:
                raise RuntimeError("There is no server public key in %s so cannot connect" % filename)
        self.server_public_binary = b64decode(server_pk)
        if len(self.server_public_binary) != 32:
            raise RuntimeError("The server public key record is broken.")

        # Threading
        zmq.Context.instance().set(zmq.IO_THREADS, psutil.cpu_count())
        logging.debug("ZeroMQ context IO threads: " + str(zmq.Context.instance().get(zmq.IO_THREADS)))

        # Prevent queued messages being sent instead of a handshake when reconnecting
        zmq.Context.instance().setsockopt(zmq.LINGER, 0)

        # Should be able to connect, then.
        self.loop_block.acquire()  # allows the thread to start and prepare but not actually run, yet.
        self.thread = Thread(target=self._start, name="Messidge background loop")
        self.thread.start()

    def start(self):
        """Start message loop - separate from __init__ so we get a chance to register_exclusive/register_commands"""
        self.loop_block.release()  # causes the block in _start to clear
        return self

    def wait_until_complete(self):
        """Blocks until the message loop exits"""
        self.thread.join()
        
        # re-raises (on the main thread) if the loop caught an exception
        if self.exception is not None:
            raise self.exception

    def disconnect(self):
        """Stop the message loop and disconnect - without this object cannot be garbage collected"""
        for skt in self.thread_skt.values():
            skt.disconnect(self.inproc_name)
        self.loop.stop()
        self.thread_skt.clear()
        self.connect_callbacks.clear()
        self.uuid_blockreply.clear()
        self.uuid_blockresults.clear()

    def _start(self):
        """The message loop runs on a background thread"""
        self.loop_thread = get_ident()

        # create the trunk socket - remember sockets must be created on the thread they are used on
        logging.info("Connecting to: %s:2020" % self.location)
        working = False
        while not working:
            try:
                self.skt = zmq.Context.instance().socket(zmq.DEALER)
                self.skt.set_hwm(0)  # infinite, don't drop or block because that would be bad
                self.skt.setsockopt(zmq.LINGER, 0)
                self.skt.connect("tcp://%s:2020" % self.connect_ip)
                self.skt_monitor = self.skt.get_monitor_socket()
                working = True
            except zmq.error.ZMQError:
                del self.skt
                self.skt = None
                logging.info("Retrying ZMQ socket")
                time.sleep(1)
        logging.debug("Trunk socket is: %x" % id(self.skt))

        # create the cross thread socket for forwarding
        # since there can be more than one connection in the same process we need to give it a unique id
        self.x_thread_receive = zmq.Context.instance().socket(zmq.DEALER)
        self.x_thread_receive.bind(self.inproc_name)
        logging.debug("Cross-thread socket is: %x" % id(self.x_thread_receive))

        # kick off a message loop - has to be constructed on background thread
        self.loop = Loop(self.skt, exit_on_exception=self.exit_on_exception)
        self.loop.register_exclusive(self.skt_monitor, self._socket_event, comment="Socket events")
        self.loop.register_exclusive(self.x_thread_receive, self._fast_forward, comment="Cross thread socket")
        if self.reflect_value_errors:
            self.loop.on_value_error(self._background_thread_exception)

        # block until allowed to go by start()
        self.loop_block.acquire()
        self.mark_as_ready()
        self.loop.run()  # blocks until loop.stop is called or the loop catches a wayward exception
        self.exception = self.loop.caught_exception

        # Maybe the main thread was in wait_until_ready
        if self.exception is not None:
            self.unblock_and_raise(self.exception)

        # Maybe the main thread was in send_blocking_cmd
        for block in self.uuid_blockreply.values():
            block.release()

    def _socket_event(self, monitor_socket):
        # is this a connection event?
        event = recv_monitor_message(monitor_socket)
        if event['event'] == zmq.EVENT_DISCONNECTED:
            if len(self.uuid_blockreply) != 0:
                raise RuntimeError("Was disconnected from location with a blocking call still pending")
            logging.info("Have been disconnected from location. Wait...")
            self.connected = False
            return
        if event['event'] != zmq.EVENT_CONNECTED:
            return

        logging.info("Message queue connected")
        self.connected = True

        # send the encryption session request
        # the rid is used to show which session we *were* if reconnecting
        params = {'user': self.keys.public_binary(), 'rid': self.rid}
        parts = [b'', b'auth', b'', params, b'']
        binary = cbor.dumps(parts)
        self.skt.send(binary)
        try:
            in_binary = self.skt.recv()
            in_parts = cbor.loads(in_binary)
            if len(in_parts) != 3:
                raise ValueError("Authentication failed.")
            recv_nonce, enc_session_key, rid = in_parts
        except ValueError as e:
            self.unblock_and_raise(e)  # raises within a blocked wait_until_ready
            return

        # unwrap and set the session key
        try:
            if enc_session_key != b'':
                self.session_key = libnacl.crypto_box_open(enc_session_key, recv_nonce,
                                                           self.server_public_binary, self.keys.secret_binary())
                self.loop.set_crypto_params(self.session_key)
        except libnacl.CryptError:
            raise ValueError("Handshake failed.")
        self.rid = rid
        logging.info("Handshake completed")
        for callback in self.connect_callbacks:
            callback(self.rid)
        self.mark_as_ready()

    def send_skt(self):
        """Allocates (if necessary) a socket for the calling thread to send messages with."""
        thread = get_ident()

        # if this is the loop thread then we can use the main thread socket to send
        if thread is self.loop_thread:
            return self.skt

        try:
            return self.thread_skt[thread]
        except KeyError:
            # need a new socket
            new_skt = zmq.Context.instance().socket(zmq.DEALER)
            new_skt.connect(self.inproc_name)
            self.thread_skt[thread] = new_skt
            logging.debug("Created a per-thread socket for: " + str(thread))
            return new_skt

    def destroy_send_skt(self):
        """Closes and removes the send_skt for the calling thread."""
        thread = get_ident()
        try:
            self.thread_skt[thread].close()
            del self.thread_skt[thread]
            logging.debug("Destroyed per-thread socket for: " + str(thread))
        except KeyError:
            pass

    def send_cmd(self, cmd: bytes, params=None, bulk: bytes=b'', uuid=b'', reply_callback=None):
        """Sends a command to the location, can route replies.

        :param cmd: the command.
        :param params: A {'key': 'value'} dictionary of parameters or ['list'].
        :param bulk: An optional piece of bulk data to transport.
        :param uuid: A uuid to attach to this command (so it can reply).
        :param reply_callback: Callback to fire when the command receives a reply - gets passed the message.
        """
        # BS check
        if self.loop is None:
            raise RuntimeError("The connection has no message loop - _start has not been called.")
        if cmd is None:
            raise RuntimeError("Need to pass a command to send_cmd.")

        # send
        if reply_callback is not None:
            # important that we register the expectation of a reply before asking the question
            if uuid is b'':
                uuid = shortuuid.uuid().encode()  # as bytes
            self.loop.register_reply(uuid, reply_callback)

        Message.send(self.send_skt(), cmd, self.session_key, params, uuid=uuid, bulk=bulk)

    def send_blocking_cmd(self, cmd: bytes, params=None, bulk: bytes=b'', timeout: float=120) -> Message:
        """Sends a command to the location and blocks waiting for a reply (which is returned).
        May raise ValueError exceptions.

        :param cmd: the command.
        :param params: A {'key': 'value'} dictionary of parameters or ['list'].
        :param bulk: An optional piece of bulk data to transport.
        :param timeout: In seconds.
        :return: The reply message.
        """
        thread = get_ident()
        if thread == self.loop_thread:
            raise RuntimeError("Cannot send a blocking command on the loop thread")

        # acquire a lock, wait for the reply
        uuid = shortuuid.uuid().encode()
        self.uuid_blockreply[uuid] = allocate_lock()
        self.uuid_blockreply[uuid].acquire()
        self.send_cmd(cmd, params, uuid=uuid, reply_callback=self._unblock, bulk=bulk)

        # when the background thread has an answer, the lock will release and we can continue
        if not self.uuid_blockreply[uuid].acquire(timeout=timeout):
            raise ValueError("Blocking call timed out: %s(%s)" % (cmd.decode(), str(params)))

        msg = self.uuid_blockresults[uuid]

        # clean up
        self.uuid_blockreply[uuid].release()
        del self.uuid_blockreply[uuid]
        del self.uuid_blockresults[uuid]

        # raise exceptions if either in the message or caught on a background thread
        if 'exception' in msg.params:
            raise ValueError(msg.params['exception'])
        if self.exception:
            raise self.exception
        return msg

    def register_commands(self, obj, commands):
        """Register a list of commands to be handled by the loop.

        :param obj: The object that will handle the commands.
        :param commands: A dict of the form {b'command': handler, ...}."""
        while self.loop is None:
            os.sched_yield()  # loop creation has not yet had a timeslice
        self.loop.register_commands(self.skt, obj, commands)

    def register_connect_callback(self, callback):
        """Register a callback to be fired once the connection is complete.

        :param callback: the object.method to call - is passed the connection rid."""
        self.connect_callbacks.add(callback)

    def unregister_connect_callback(self, callback):
        """Unregister a callback to be fired once the connection is complete.

        :param callback: the object.method to cancel."""
        try:  # may have already been disconnected as part of object shutting down
            self.connect_callbacks.remove(callback)
        except KeyError:
            pass

    def location_name(self) -> str:
        """Outside access to the name/address of this location.

        :return: The FQDN of the location as a string."""
        return self.location

    def _fast_forward(self, skt):
        binary = skt.recv()
        self.skt.send(binary)

    def _unblock(self, msg: Message):
        try:
            self.uuid_blockresults[msg.uuid] = msg  # msg is stored first
            self.uuid_blockreply[msg.uuid].release()
            self.loop.unregister_reply(msg.uuid)
        except KeyError:
            logging.error("Message from a blocking call arrived late (ignored): " + str(msg))

    def _background_thread_exception(self, e, msg):
        """Call from non-main threads to return a value error to the client"""
        if msg.replyable():
            msg.reply(self.send_skt(), {"exception": str(e)})
            logging.info("Client called %s and raised a ValueError: %s" % (str(msg.command), str(e)))
        else:
            logging.warning("Unable to return ValueError to client: " + str(e))

    def __repr__(self):
        return "<messidge.client.connection.Connection object at %s (location=%s)>" % (id(self), self.location)


def cmd(required_params, *, needs_reply=False):
    """Create the internal structure describing a command

    :param required_params: A list of parameters that must be included with the command.
    :param needs_reply: The message needs to be replied to (and must have a uuid)."""
    return required_params, needs_reply



# Copyright (c) 2017 David Preece - davep@polymath.tech, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# Base classes for models and controllers

from lru import LRU


class NodeMinimal:
    def __init__(self, pk, msg, config):
        self.pk = pk
        self.msg = msg
        self.config = config


class SessionMinimal:
    def __init__(self, rid, pk):
        self.rid = rid
        self.pk = pk
        self.old_rid = rid  # for use with reconnection

    # overload to free resources - can use passed broker to send commands to nodes
    def close(self, broker):
        pass


class ModelMinimal:
    """Stores the nodes and user sessions attached to this broker."""

    def __init__(self):
        self.nodes = {}
        self.sessions = {}
        self.forward_replies = LRU(2048)

    def resources(self, pk):
        """Overload this method to return a resource offer to a newly connected client.

        :param pk: the public key of the connecting user. """
        return None

    # overload these if you want to make persistent sessions
    def create_session_record(self, sess):
        pass

    def update_session_record(self, sess):
        pass

    def delete_session_record(self, sess):
        pass



# Copyright (c) 2017 David Preece - davep@polymath.tech, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.


from setuptools import setup

setup(
    name='messidge',
    version='1.1.2',
    author='David Preece',
    author_email='davep@20ft.nz',
    url='https://20ft.nz',
    license='BSD',
    packages=['messidge', 'messidge.broker', 'messidge.client'],
    install_requires=['pyzmq', 'libnacl', 'shortuuid', 'psutil', 'lru-dict', 'cbor', 'bottle', 'litecache'],
    description='A message passing library for gateways onto the public Internet.',
    long_description="There are some users, and some nodes (on a private LAN). Messidge brokers between the two. "
                     "It authenticates, encrypts, and can validate the provenance of an rpc call.",
    keywords='message messages message-passing authenticated encrypted gateway',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Framework :: Bottle',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: BSD License',
        'Topic :: Security',
        'Topic :: Security :: Cryptography',
        'Topic :: Software Development :: Object Brokering',
        'Topic :: System :: Distributed Computing',
        'Topic :: System :: Networking',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'
    ]
)



# Copyright (c) 2017 David Preece - davep@polymath.tech, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# uuid is used to identify replies to commands including long running replies (with more than one reply)
# identifiers for everything else are to be passed explicitly in the params (as strings, because of json)

import cbor
import libnacl
import libnacl.utils
import logging


class Message:
    def __init__(self, session_key=None):
        self.command = None
        self.uuid = None
        self.params = None
        self.bulk = None
        self.session_key = session_key  # here so we can reply to a message without knowledge of the session itself

    @staticmethod
    def receive(socket, session_key):
        """Pulls one message off the socket, decrypts and returns it.

        :param socket: The ZMQ socket to receive from.
        :param session_key: The session key for this connection.
        :return: The received and decrypted message."""
        # parts are command, uuid, params, bulk
        rtn = Message(session_key)
        binary = socket.recv()
        parts = cbor.loads(binary)
        nonce = bytes(parts[0])
        rtn.command = bytes(parts[1])
        rtn.uuid = bytes(parts[2])
        rtn.params, rtn.bulk = Message.decrypted_params(parts[3],
                                                        parts[4],
                                                        nonce,
                                                        session_key)
        
        logging.debug("Message.receive (%s:%s:%s)" %
                      (rtn.uuid.decode(),
                       rtn.command.decode(),
                       list(rtn.params.keys()) if isinstance(rtn.params, dict) else "--"))
        return rtn

    @staticmethod
    def from_persist(parts):
        """Creates a message from four previously persisted parts.

        :param parts: The four parts.
        :return: The constructed message."""
        rtn = Message()
        rtn.command = parts[0]
        rtn.uuid = parts[1]
        rtn.params = parts[2]
        rtn.bulk = parts[3]
        return rtn

    def for_persist(self) -> []:
        """Create a list of four parts that can be used to recreate this message.

        :return: A four element list that can be used to persist the message."""
        return [self.command, self.uuid, self.params, self.bulk]

    def replyable(self) -> bool:
        """Can this message be replied to?

        :return: True/False"""
        return self.uuid != b''

    def reply(self, socket, results=None, bulk=b''):
        """Reply to a previously received message.

        :param socket: the socket to use to send the message.
        :param results: an optional dictionary of results to include as part of the reply.
        :param bulk: an optional bulk data blob."""
        if not self.replyable():
            logging.warning("Tried to reply to a non-replyable message: " + str(self))
            return

        logging.debug("Message.reply: (%s:%s:%s)" %
                      (self.uuid.decode(),
                       self.command.decode(),
                       list(results.keys()) if isinstance(results, dict) else "--"))
        if results is None:
            results = {}
        Message.send(socket, b'', self.session_key, results, uuid=self.uuid, bulk=bulk, trace=False)

    def forward(self, socket):
        """Forward a message unchanged through another socket.

        :param socket: the socket to use to forward the message."""
        Message.send(socket, self.command, self.session_key, self.params, uuid=self.uuid, bulk=self.bulk)

    @staticmethod
    def send(socket, command, session_key, params=None, *, uuid=b'', bulk=b'', trace=True):
        """Send a command to the location. Can be called directly but much better to use Connection.send_cmd.

        :param socket: socket to use to send the message.
        :param command: the command as a binary string.
        :param session_key: The session key for this connection.
        :param params: An optional {'key': 'value'} dictionary of parameters or ['list'].
        :param uuid: An optional uuid to attach to the message. This is what makes messages replyable.
        :param bulk: An optional piece of bulk data to transport.
        :param trace: Set to false to disable debug logging on this occasion.
        """
        if trace:
            logging.debug("Message.send: (%s:%s:%s)" %
                          (uuid.decode(),
                           command.decode(),
                           list(params.keys()) if isinstance(params, dict) else "--"))
        if params is None:
            params = {}
        nonce = libnacl.utils.rand_nonce()
        params_encrypted, bulk_encrypted = Message.encrypted_params(params, bulk, nonce, session_key)
        parts = [nonce,
                 command,
                 uuid,
                 params_encrypted,
                 bulk_encrypted]
        binary = cbor.dumps(parts)
        socket.send(binary)

    @staticmethod
    def encrypted_params(params, bulk, nonce, session_key) -> (bytes, bytes):
        """Returns the encrypted form of the passed parameter dictionary and bulk data.

        :param params: The parameter dictionary.
        :param bulk: The bulk data.
        :param nonce: The n-once for this message.
        :param session_key: The session key for this connection.
        :return: A (bytes, bytes) tuple for the encrypted params and bulk data.
        """
        if session_key is not None:
            params_encrypted = libnacl.crypto_secretbox(cbor.dumps(params), nonce, session_key)
            bulk_encrypted = libnacl.crypto_secretbox(bulk, nonce, session_key) if bulk != b'' else b''
            return params_encrypted, bulk_encrypted
        else:
            return params, bulk

    @staticmethod
    def decrypted_params(params, bulk, nonce, session_key):
        """Returns the decrypted form of the passed parameter dictionary and bulk data.

        :param params: The parameter dictionary.
        :param bulk: The bulk data.
        :param nonce: The n-once for this message.
        :param session_key: The session key for this connection.
        :return: A (bytes, bytes) tuple for the decrypted params and bulk data.
        """
        if session_key is not None:
            params_plaintext = libnacl.crypto_secretbox_open(params, nonce, session_key)
            bulk_plaintext = libnacl.crypto_secretbox_open(bulk, nonce, session_key) if bulk != b'' else b''
            return cbor.loads(params_plaintext), bulk_plaintext
        else:
            return params, bulk

    def __repr__(self):
        return "<messidge.message.Message object at %x (command=%s uuid=%s params=%s)>" % \
               (id(self), self.command, str(self.uuid), self.params)
